// Activator.cpp : implementation file
//

#include "stdafx.h"
#include "Activator.h"

#include "..\SOAPObjectLib\SOAPEnvelope.h"
#include "..\SOAPObjectLib\SOAPBody.h"
#include "..\SOAPObjectLib\SOAPIndElement.h"
#include "..\SOAPObjectLib\SOAPEmbeddedElement.h"
#include "..\SOAPObjectLib\SOAPHttpPacket.h"

#include <atlbase.h>

#include <strstream>
#include <string>

// Constructor
CActivator::CActivator(BSTR bstrURI, BSTR bstrMethod) : 
    m_bstrMethodURI(bstrURI),
    m_bstrMethod(bstrMethod)
{
}

//
// Invoke the method call based on the incoming XML
// request and generate an XML response from the result.
//
void CActivator::Invoke(IXMLDOMDocument *pRequest,
					   IXMLDOMDocument **ppResponse)
{
	try {
		// Build outgoing XML (caller is responsible for
		// releasing this interface)
		HRESULT hr = CoCreateInstance(__uuidof(DOMDocument), 
							  0, 
							  CLSCTX_ALL, 
							  __uuidof(IXMLDOMDocument), 
							  reinterpret_cast<void **>(ppResponse));
		if ( FAILED(hr) ) throw hr;

		// Check the HTTP SOAPAction header to see if the name of
		// the method there reflects the method we find in the 
		// body.
		//
 		// Check for <SOAP-ENV:Envelope>
		//              <SOAP-ENV:Body>
		//
		// Construct search string "SOAP-ENV:Envelope/SOAP-ENV:Body"
        CComBSTR bstrSS(L"SOAP-ENV:Envelope/SOAP-ENV:Body");

        // Find the body
		CComPtr<IXMLDOMNode> spBodyNode;
        hr = pRequest->selectSingleNode(bstrSS,&spBodyNode);
		if ( FAILED(hr) ) throw hr;

        // The method must be the first child of the body
		CComPtr<IXMLDOMNode> spMethodNode;
        hr = spBodyNode->get_firstChild(&spMethodNode);
		if ( FAILED(hr) ) throw hr;

        // Ensure that the HTTP method name matches the payload
		// method name.
        CComBSTR bstrPayloadMethodName;
        hr = spMethodNode->get_baseName(&bstrPayloadMethodName);
        if ( wcscmp(bstrPayloadMethodName,m_bstrMethod) ) {
            // Wrong method, or at least the packet information
			// doesn't agree.  The method name in the SOAPAction
			// header should match what's in the XML payload.
			throw E_NOTIMPL;
        } // if

		// Check for supported method name "GetRemoteTime"
		std::string strReturn;
		char szData[33] = {0};
		if ( !_wcsicmp(L"GetRemoteTime",m_bstrMethod) ) {
			// Obtain the server's system time
			SYSTEMTIME st;
			GetSystemTime(&st);

			// Format: 2000-02-01T14:05:00
			// Note the server will return its time as
			// UTC, so no bias will be required...
			//
			// Convert year first
			itoa(st.wYear,szData,10);
			strReturn = szData;
			strReturn += "-";

			// Convert month
			itoa(st.wMonth,szData,10);
			strReturn += szData;
			strReturn += "-";

			// Convert day
			itoa(st.wDay,szData,10);
			strReturn += szData;
			strReturn += "T";

			// Convert hour
			//
			// Note SYSTEMTIME is in UTC,
			// so no bias is required
			itoa(st.wHour,szData,10);
			strReturn += szData;
			strReturn += ":";

			// Convert minute
			itoa(st.wMinute,szData,10);
			strReturn += szData;
			strReturn += ":";

			// Convert second
			itoa(st.wSecond,szData,10);
			strReturn += szData;
		} // if
		else {
			// Fault...we don't support this method
			throw E_NOTIMPL;
		} // else

		// Create response packet
		CreateResponse((*ppResponse),strReturn);
	} // try
	catch(HRESULT hrErr) {
		// Create fault packet
		if ( (*ppResponse) != NULL ) {
			CreateFault((*ppResponse),hrErr);
		} // if
	} // catch
	catch(...) {
		// Create fault packet
		if ( (*ppResponse) != NULL ) {
			CreateFault((*ppResponse),E_UNEXPECTED);
		} // if
	} // catch
}

void CActivator::CreateResponse(IXMLDOMDocument *pResponse,
								const std::string& strReturn)
{
	// Build a successful SOAP response packet
	CSOAPHttpPacket* pPacket = NULL;
	try {
		// Build SOAP HTTP packet (only used as a container
		// for the Envelope)
		SOAPCALLINFO ciInfo;
		ciInfo.strHost    = "none";
		ciInfo.strPostHdr = "none";
		ciInfo.strMethod  = "none"; // won't be used for return packet
		pPacket = new CSOAPHttpPacket(&ciInfo);
		if ( pPacket == NULL ) throw E_OUTOFMEMORY;

		// Build SOAP:Envelope
		CSOAPEnvelope* pEnvelope = pPacket->GetEnvelope();
		long iCookie = NULLNSCOOKIE;
		USES_CONVERSION;
		HRESULT hr = pEnvelope->InsertNamespace(std::string(W2A(m_bstrMethodURI)), 
												std::string("m"), 
												&iCookie);
		if ( FAILED(hr) ) throw hr;

		// Build SOAP:Body and SOAP:MethodResponse
   		CSOAPBody* pBody = pEnvelope->GetBody();
		CSOAPIndElement* pMethodElement = NULL;
		std::string strMethodResponse(std::string(W2A(m_bstrMethod) + std::string("Response")));
		hr = pBody->InsertElement(strMethodResponse, iCookie, &pMethodElement);
		if ( FAILED(hr) ) throw hr;

		// Build SOAP:MethodResponse <return> param
		// (which is the system time string)
		CSOAPEmbeddedElement* pEmbedded = NULL;
		hr = pMethodElement->InsertElement(std::string("return"), NULLNSCOOKIE, &pEmbedded);
		if ( FAILED(hr) ) throw hr;
		pEmbedded->SetElementData(strReturn);

		// Write XML Response payload to stream (without HTTP headers)
		std::ostrstream ostr;
		hr = pPacket->WriteToStream(ostr, true);

		// Get XML (SOAP) document
		CComBSTR bstrResponse(ostr.pcount()+1,ostr.str());
		ostr.freeze(0); // allows deletion of character stream upon destruction

		// Place XML in DOM
		VARIANT_BOOL bSuccess = false;
		hr = pResponse->loadXML(bstrResponse,&bSuccess);
		if ( hr == S_FALSE ) {
			// Assume memory problem...if this is the case,
			// then we won't be able to send a fault back
			// either.  We'll deal with that issue in the
			// catch block.
			throw E_OUTOFMEMORY;
		} // if
	} // try
	catch(...) {
		// Create fault packet...this may have happened because
		// we couldn't load the XML document into the parser,
		// so this call will also result in failure.  If so,
		// the returned XML DOM object will have no XML
		// document, and the caller will have to assume something
		// catastrophic happened...
		CreateFault(pResponse,E_UNEXPECTED);
	}

	// Clean up...
	if ( pPacket != NULL ) delete pPacket;
}


void CActivator::CreateFault(IXMLDOMDocument *pResponse,
							 HRESULT hrErr)
{
	// Build a successful SOAP response packet
	CSOAPHttpPacket* pPacket = NULL;
	try {
		// Check for the HRESULT and insert an appropriate
		// fault message
		std::string strFaultCode, strFaultString;
		switch ( hrErr ) {
			case E_NOTIMPL:
				strFaultCode = "Client.NotImplemented";
				strFaultString = "The requested method is not implemented";
				break;

			case E_OUTOFMEMORY:
				strFaultCode = "Client.OutOfMemory";
				strFaultString = "There is not enough server memory available to execute your request";
				break;

			case E_INVALIDARG:
				strFaultCode = "Client.InvalidArgument";
				strFaultString = "You provided an invalid method argument";
				break;

			case E_UNEXPECTED:
				strFaultCode = "Client.UnexpectedError";
				strFaultString = "Something unexpected and not entirely delightful resulted from your request";
				break;

			default:
				strFaultCode = "Client.UnknownError";
				strFaultString = "Call your broker and dump all EnduraSoft stock immediately";
				break;

		} // switch

		// Build SOAP HTTP packet (only used as a container
		// for the Envelope)
		SOAPCALLINFO ciInfo;
		ciInfo.strHost    = "none";
		ciInfo.strPostHdr = "none";
		ciInfo.strMethod  = "none"; // won't be used for return packet
		pPacket = new CSOAPHttpPacket(&ciInfo);
		if ( pPacket == NULL ) throw E_OUTOFMEMORY;

		// Build SOAP:Envelope
		CSOAPEnvelope* pEnvelope = pPacket->GetEnvelope();

		// Build SOAP:Body and SOAP:Fault
   		CSOAPBody* pBody = pEnvelope->GetBody();
		CSOAPIndElement* pFaultElement = NULL;
		HRESULT hr = pBody->InsertElement(std::string("Fault"), ENVNSCOOKIE, &pFaultElement);
		if ( FAILED(hr) ) throw hr;

		// Build SOAP:Fault return information, starting with
		// </faultcode>
		CSOAPEmbeddedElement* pFaultCode = NULL;
		hr = pFaultElement->InsertElement(std::string("faultcode"), NULLNSCOOKIE, &pFaultCode);
		if ( FAILED(hr) ) throw hr;
		pFaultCode->SetElementData(strFaultCode);

		// </faultstring>
		CSOAPEmbeddedElement* pFaultString = NULL;
		hr = pFaultElement->InsertElement(std::string("faultstring"), NULLNSCOOKIE, &pFaultString);
		if ( FAILED(hr) ) throw hr;
		pFaultString->SetElementData(strFaultString);

		// Write XML Response payload to stream (without HTTP headers)
		std::ostrstream ostr;
		hr = pPacket->WriteToStream(ostr, true);

		// Null terminate the XML
		std::string str = ostr.str();
		str[ostr.pcount()] = '\0';
		ostr.freeze(0); // allows deletion of character stream upon destruction

		// Place XML in DOM
		VARIANT_BOOL bSuccess = false;
		hr = pResponse->loadXML(CComBSTR(str.c_str()), &bSuccess);
		if ( hr == S_FALSE ) {
			// Assume memory problem...if this is the case,
			// then we won't be able to send a fault back
			// either.  We'll deal with that issue in the
			// catch block.
			throw E_OUTOFMEMORY;
		} // if
	} // try
	catch(...) {
		// If we get here, something really bad happened, so
		// clear the XML DOM object and return it empty.  The
		// caller is responsible for interpreting the result
		// of this action, which you can assume to be 
		// catastrophic.  Calling the CComBSTR constructor
		// simply creates a NULL BSTR, which clears the
		// XML DOM object's document.
		VARIANT_BOOL bSuccess = false;
		pResponse->loadXML(CComBSTR(), &bSuccess);
	} // catch

	// Clean up...
	if ( pPacket != NULL ) delete pPacket;
}