// SOAPElement.cpp
//
// Kenn Scribner
// Understanding SOAP, March 2000
//
// Basic SOAP XML element object definition.
// Implements the basic SOAP (XML) node and acts
// as the base class for all derived SOAP elements
// (envelope, header, body, independent, and
// embedded).
//

// Disable the "truncated to 255 characters" warning
// when expanding STL templates...
#pragma warning(disable : 4786)

#include "SOAPElement.h"

CSOAPElement::CSOAPElement(CSOAPElement* pParent /*=NULL*/) :
  m_pParent(pParent),
  m_iTagNamespaceCookie(NULLNSCOOKIE)
{
	// Establish attribute/namespace scoping
	if ( m_pParent != NULL ) {
		// Establish attribute namespace scoping
		m_CAttributes.SetParent(m_pParent);

		// Establish namespace scoping
		m_CNamespaces.SetParent(m_pParent);
	} // if
}

CSOAPElement::~CSOAPElement()
{
}

void CSOAPElement::SetTagText(const std::string& strTagText)
{
	// Copy provided string
	m_strTagText = strTagText;
}

void CSOAPElement::SetTagNamespace(long iNamespaceCookie)
{
	// Store provided cookie
	m_iTagNamespaceCookie = iNamespaceCookie;
}

bool CSOAPElement::EnableStdAttribute(long iStdAttributeCookie, bool bEnable)
{
	// Forward to attribute object
	return m_CAttributes.EnableStdAttribute(iStdAttributeCookie,bEnable);
}

unsigned long CSOAPElement::InsertAttribute(long iNamespaceCookie, std::string& strAttrName, std::string& strAttrValue)
{
	// Forward to attribute object
	return m_CAttributes.InsertAttribute(iNamespaceCookie,strAttrName,strAttrValue);
}

void CSOAPElement::SetElementData(const std::string& strData, bool bIsCDATA /*=FALSE*/)
{
	// Copy provided string
	if ( bIsCDATA ) {
		// Wrap the text in a CDATA block
		// TODO: Should check incoming string for "]]"!
		m_strData = "<![CDATA[";
		m_strData += strData;
		m_strData += "]]>";
	} // if
	else {
		// Copy text as text
		m_strData = strData;
	} // else
}

bool CSOAPElement::EnableStdNamespace(long iStdNamespaceCookie, bool bEnable)
{
	// Forward to namespace object
	return m_CNamespaces.EnableStdNamespace(iStdNamespaceCookie,bEnable);
}

unsigned long CSOAPElement::InsertNamespace(const std::string& strLongName, const std::string& strShortName, long* piNamespaceCookie)
{
	// Delegate call...
	return m_CNamespaces.InsertNamespace(strLongName,strShortName,piNamespaceCookie);
}

unsigned long CSOAPElement::GetNamespace(long iNamespaceCookie, std::string& strNamespace)
{
	// Delegate call...
	return m_CNamespaces.GetNamespace(iNamespaceCookie,strNamespace);
}

unsigned long CSOAPElement::WriteToStream(std::ostrstream& ostr)
{
	// Nothing to write if the external stream is NULL
	if ( ostr == NULL ) return SOAP_E_POINTER;

	unsigned long res = SOAP_S_OK;
	try {
		// Write the opening tag
		res = StreamOpenTag(ostr);
		if ( res ) throw res;

		// Write the contents
		res = StreamContents(ostr);
		if ( res ) throw res;

		// Write the closing tag
		res = StreamCloseTag(ostr);
		if ( res ) throw res;
	} // try
	catch(unsigned long resErr) {
		// Some internal error...
		res = resErr; // normalize to this error...
		char szErr[256] = {0};
		sprintf(szErr,"Internal error while streaming element, %#08x.",resErr);
		std::cerr << szErr << std::endl;
	} // catch
	catch(...) {
		// Some error...
		res = SOAP_E_UNEXPECTED;
		std::cerr << "Unexpected application error while streaming element" << std::endl;
	} // catch

	return res;
}

unsigned long CSOAPElement::StreamContents(std::ostrstream& ostr)
{
	// Nothing to write if the external stream is NULL
	if ( ostr == NULL ) return SOAP_E_POINTER;

	unsigned long res = SOAP_S_OK;
	try {
		// Write the contents
		ostr << m_strData.c_str();
	} // try
	catch(unsigned long resErr) {
		// Some internal error...
		res = resErr; // normalize to this error...
		char szErr[256] = {0};
		sprintf(szErr,"Internal error while streaming element contents, %#08x.",resErr);
		std::cerr << szErr << std::endl;
	} // catch
	catch(...) {
		// Some error...
		res = SOAP_E_UNEXPECTED;
		std::cerr << "Unexpected application error while streaming element contents" << std::endl;
	} // catch

	return res;
}

unsigned long CSOAPElement::StreamOpenTag(std::ostrstream& ostr)
{
	// Nothing to write if the external stream is NULL
	if ( ostr == NULL ) return SOAP_E_POINTER;

	unsigned long res = SOAP_S_OK;
	try {
		// Write the opening tag text to the local stream,
		// if the tag is named
		if ( m_strTagText.length() ) {
			std::string sTag("<");

			// Stream namespace, if any
			if ( m_iTagNamespaceCookie != NULLNSCOOKIE) {
				std::string strNamespace;
				GetNamespace(m_iTagNamespaceCookie,strNamespace);
				sTag += strNamespace;

				// Concatenate the colon
				sTag += ":";
			} // if

			// Concatenate the tag's element name
			sTag += m_strTagText;

			// Write the tag prior to attribute insertion
			ostr << sTag.c_str();

			// Stream any namespaces
			res = m_CNamespaces.WriteToStream(ostr);
			if ( res ) throw res;

			// Stream any attributes
			res = m_CAttributes.WriteToStream(ostr);
			if ( res ) throw res;

			// Terminate the tag
			if ( m_strData.length() ) {
				// Normal tag termination as this element has data
				sTag = ">";
			} // if
			else {
				// Empty element tag termination, as this element has
				// no data (such as an embedded multi-reference
				// XPointer element in the method independent element).
				sTag = " />";
			} // else

			// Close the tag
			ostr << sTag.c_str();
		} // if
	} // try
	catch(unsigned long resErr) {
		// Some internal error...
		res = resErr; // normalize to this error...
		char szErr[256] = {0};
		sprintf(szErr,"Internal error while streaming element opening tag, %#08x.",resErr);
		std::cerr << szErr << std::endl;
	} // catch
	catch(...) {
		// Some error...
		res = SOAP_E_UNEXPECTED;
		std::cerr << "Unexpected application error while streaming element opening tag" << std::endl;
	} // catch

	return res;
}

unsigned long CSOAPElement::StreamCloseTag(std::ostrstream& ostr)
{
	// Nothing to write if the external stream is NULL
	if ( ostr == NULL ) return SOAP_E_POINTER;

	unsigned long res = SOAP_S_OK;
	try {
		// Write the closing tag text to the local stream,
		// if there was data to enclose
		if ( m_strData.length() ) {
			std::string sTag("</");

			// Stream namespace, if any
			if ( m_iTagNamespaceCookie != NULLNSCOOKIE) {
				std::string strNamespace;
				GetNamespace(m_iTagNamespaceCookie,strNamespace);
				sTag += strNamespace;

				// Concatenate the colon
				sTag += ":";
			} // if

			// Concatenate the tag's element name
			sTag += m_strTagText;

			// Terminate the tag
			sTag += ">";

			// Write the tag
			ostr << sTag.c_str();
		} // if
	} // try
	catch(unsigned long resErr) {
		// Some internal error...
		res = resErr; // normalize to this error...
		char szErr[256] = {0};
		sprintf(szErr,"Internal error while streaming element closing tag, %#08x.",resErr);
		std::cerr << szErr << std::endl;
	} // catch
	catch(...) {
		// Some error...
		res = SOAP_E_UNEXPECTED;
		std::cerr << "Unexpected application error while streaming element closing tag" << std::endl;
	} // catch

	return res;
}

