// SOAPElementAttributes.cpp
//
// Kenn Scribner
// Understanding SOAP, March 2000
//
// SOAP XML element attribute object implementation
//

// Disable the "truncated to 255 characters" warning
// when expanding STL templates...
#pragma warning(disable : 4786)

#include "SOAPElement.h"
#include "SOAPElementAttributes.h"

CSOAPElementAttributes::CSOAPElementAttributes(CSOAPElement* pParent /*=NULL*/) :
  m_pParent(pParent)
{
	// Clear the standard attributes
	m_bStdAttributes[ENCATTRCOOKIE] = false; // encodingStyle
	m_bStdAttributes[MUNDATTRCOOKIE] = false; // mustUnderstand
	m_bStdAttributes[ACTNXTATTRCOOKIE] = false; // actor
	m_bStdAttributes[ROOTATTRCOOKIE] = false; // root
}

CSOAPElementAttributes::~CSOAPElementAttributes()
{
}

bool CSOAPElementAttributes::EnableStdAttribute(long iStdAttributeCookie, bool bEnable)
{
	// Check for global enable/disable
	bool bPrevious = false;
	if ( iStdAttributeCookie == GLOBATTRCOOKIE ) {
		// Globally enable/disable attributes
		m_bStdAttributes[ENCATTRCOOKIE] = bEnable; // encodingStyle
		m_bStdAttributes[MUNDATTRCOOKIE] = bEnable; // mustUnderstand
		m_bStdAttributes[ACTNXTATTRCOOKIE] = bEnable; // actor
		m_bStdAttributes[ROOTATTRCOOKIE] = bEnable; // root
	} // if
	else {
		// Enable/disable the specific standard attribute
		switch ( iStdAttributeCookie ) {
			case ENCATTRCOOKIE:
			case MUNDATTRCOOKIE:
			case ACTNXTATTRCOOKIE:
			case ROOTATTRCOOKIE:
				bPrevious = m_bStdAttributes[iStdAttributeCookie];
				m_bStdAttributes[iStdAttributeCookie] = bEnable;
				break;

			default:
				break;
		} // switch
	} // else

	return bPrevious;
}

unsigned long CSOAPElementAttributes::InsertAttribute(long iNamespaceCookie, std::string& strAttrName, std::string& strAttrValue)
{
	// First find the namespace
	std::string strNamespace;
	unsigned long res = SOAP_S_OK;
	if ( m_pParent != NULL ) {
		res = m_pParent->GetNamespace(iNamespaceCookie,strNamespace);
		if ( res ) return res;
	} // if

	// Create the attribute, starting with the namespace
	std::string strAttr(" "); // leading space

	// Add the namespace, if any
	if ( strNamespace.length() ) {
		strAttr += strNamespace;

		// Add the colon
		strAttr += ":";
	} // if

	// Add the name
	strAttr += strAttrName;

	// Add the '="'
	strAttr += "=\"";

	// Add the value
	strAttr += strAttrValue;

	// Add the final quote
	strAttr += "\"";

	// Stuff in the storage vector
	m_vecAttributes.push_back(strAttr);

	return res;
}

unsigned long CSOAPElementAttributes::StreamStdAttribute(std::ostrstream& ostr, long iStdAttributeCookie)
{
	// First find the namespace
	std::string strNamespace;
	unsigned long res = SOAP_S_OK;
	if ( m_pParent != NULL ) {
		res = m_pParent->GetNamespace(ENVNSCOOKIE,strNamespace);
		if ( res ) return res;
	} // if

	// Create the attribute, starting with the namespace
	std::string strAttr(" "); // leading space

	// Add the namespace, if any
	if ( strNamespace.length() ) {
		strAttr += strNamespace;

		// Add the colon
		strAttr += ":";
	} // if

	// Stream the selected cookie
	switch ( iStdAttributeCookie ) {
		case ENCATTRCOOKIE:
			// Add the name
			strAttr += "encodingStyle";

			// Add the '="'
			strAttr += "=\"";

			// Add the value
			strAttr += "http://schemas.xmlsoap.org/soap/encoding/";

			// Add the final quote
			strAttr += "\"";
			break;

		case MUNDATTRCOOKIE:
			// Add the name
			strAttr += "mustUnderstand";

			// Add the '="'
			strAttr += "=\"";

			// Add the value
			strAttr += "1";

			// Add the final quote
			strAttr += "\"";
			break;

		case ACTNXTATTRCOOKIE:
			// Add the name
			strAttr += "actor";

			// Add the '="'
			strAttr += "=\"";

			// Add the value
			strAttr += "http://schemas.xmlsoap.org/soap/actor/next";

			// Add the final quote
			strAttr += "\"";
			break;

		case ROOTATTRCOOKIE:
			// Add the name
			strAttr += "root";

			// Add the '="'
			strAttr += "=\"";

			// Add the value
			strAttr += "1";

			// Add the final quote
			strAttr += "\"";
			break;

		default:
			res = SOAP_E_INVALIDARG;
			break;
	} // switch

	// If successful so far, write the attribute...
	if ( !res ) {
		ostr << strAttr.c_str();
	} // if

	return res;
}

unsigned long CSOAPElementAttributes::WriteToStream(std::ostrstream& ostr)
{
	// Nothing to write if the external stream is NULL
	if ( ostr == NULL ) return SOAP_E_POINTER;

	unsigned long res = SOAP_S_OK;
	try {
		// Write the standard attributes
		if ( m_bStdAttributes[ENCATTRCOOKIE] ) {
			// Stream the encodingStyle attribute
			res = StreamStdAttribute(ostr,ENCATTRCOOKIE);
			if ( res ) throw res;
		} // if

		if ( m_bStdAttributes[MUNDATTRCOOKIE] ) {
			// Stream the mustUnderstand attribute
			res = StreamStdAttribute(ostr,MUNDATTRCOOKIE);
			if ( res ) throw res;
		} // if

		if ( m_bStdAttributes[ACTNXTATTRCOOKIE] ) {
			// Stream the actor attribute
			res = StreamStdAttribute(ostr,ACTNXTATTRCOOKIE);
			if ( res ) throw res;
		} // if

		if ( m_bStdAttributes[ROOTATTRCOOKIE] ) {
			// Stream the actor attribute
			res = StreamStdAttribute(ostr,ROOTATTRCOOKIE);
			if ( res ) throw res;
		} // if

		// Write any instance-specific attributes
		for ( long i = 0; i < m_vecAttributes.size() ; i++ ) {
			// Retrieve the element
			std::string strAttr = m_vecAttributes[i];

			// Write the attribute
			ostr << strAttr.c_str();
		} // for
	} // try
	catch(unsigned long resErr) {
		// Some internal error...
		res = resErr; // normalize to this error...
		char szErr[256] = {0};
		sprintf(szErr,"Internal error while streaming element attribute, %#08x.",resErr);
		std::cerr << szErr << std::endl;
	} // catch
	catch(...) {
		// Some error...
		res = SOAP_E_UNEXPECTED;
		std::cerr << "Unexpected application error while streaming element attribute" << std::endl;
	} // catch

	return res;
}
