// SOAPElementNamespace.cpp
//
// Kenn Scribner
// Understanding SOAP, March 2000
//
// SOAP XML element namespace object implementation
//

// Disable the "truncated to 255 characters" warning
// when expanding STL templates...
#pragma warning(disable : 4786)

#include "SOAPElement.h"
#include "SOAPElementNamespace.h"

CSOAPElementNamespace::CSOAPElementNamespace(CSOAPElement* pParent /*=NULL*/, long iBaseNSCookie /*=NULLNSCOOKIE*/) :
  m_pParent(pParent),
  m_iBaseNSCookie(iBaseNSCookie),
  m_iNextCookie(0)
{
	// Insert the standard namespaces into the maps
	m_mapNSFull[ENVNSCOOKIE] = "http://schemas.xmlsoap.org/soap/envelope/";
	m_mapNSShort[ENVNSCOOKIE] = "SOAP-ENV";
	m_bStdNamespaces[ENVNSCOOKIE] = false;
	++m_iNextCookie;
	m_mapNSFull[ENCNSCOOKIE] = "http://schemas.xmlsoap.org/soap/encoding/";
	m_mapNSShort[ENCNSCOOKIE] = "SOAP-ENC";
	m_bStdNamespaces[ENCNSCOOKIE] = false;
	++m_iNextCookie;
	m_mapNSFull[XSINSCOOKIE] = "http://www.w3.org/1999/XMLSchema/instance";
	m_mapNSShort[XSINSCOOKIE] = "xsi";
	m_bStdNamespaces[XSINSCOOKIE] = false;
	++m_iNextCookie;
	m_mapNSFull[COMNSCOOKIE] = "urn:www-microsoft-com:com";
	m_mapNSShort[COMNSCOOKIE] = "COM";
	m_bStdNamespaces[COMNSCOOKIE] = false;
	++m_iNextCookie;
}

CSOAPElementNamespace::~CSOAPElementNamespace()
{
}

bool CSOAPElementNamespace::EnableStdNamespace(long iNamespaceCookie, bool bEnable)
{
	// Check for global enable/disable
	bool bPrevious = false;
	if ( iNamespaceCookie == NULLNSCOOKIE ) {
		// Globally enable/disable attributes
		m_bStdNamespaces[ENVNSCOOKIE] = bEnable;
		m_bStdNamespaces[ENCNSCOOKIE] = bEnable;
		m_bStdNamespaces[XSINSCOOKIE] = bEnable;
		m_bStdNamespaces[COMNSCOOKIE] = bEnable;
	} // if
	else {
		// Enable/disable the specific standard namespace
		switch ( iNamespaceCookie ) {
			case ENVNSCOOKIE:
			case ENCNSCOOKIE:
			case XSINSCOOKIE:
			case COMNSCOOKIE:
				bPrevious = m_bStdNamespaces[iNamespaceCookie];
				m_bStdNamespaces[iNamespaceCookie] = bEnable;
				break;

			default:
				break;
		} // switch
	} // else

	return bPrevious;
}

unsigned long CSOAPElementNamespace::InsertNamespace(const std::string& strLongName, const std::string& strShortName, long* piNamespaceCookie)
{
	// A quick check
	if ( piNamespaceCookie == NULL ) return SOAP_E_POINTER;

	// Insert the namespace
	m_mapNSFull[m_iNextCookie] = strLongName;
	m_mapNSShort[m_iNextCookie] = strShortName;

	// Assign the cookie
	*piNamespaceCookie = (m_iNextCookie | m_iBaseNSCookie);
	++m_iNextCookie;

	return SOAP_S_OK;
}

unsigned long CSOAPElementNamespace::GetNamespace(long iNamespaceCookie, std::string& strNamespace)
{
	// Look up namespace given the cookie
	unsigned long res = SOAP_S_OK;

	if ( iNamespaceCookie != NULLNSCOOKIE ) {
		// The cookie isn't asking for a null namespace, so
		// now check for the default SOAP namespace case
		if ((iNamespaceCookie >= ENVNSCOOKIE) && (iNamespaceCookie <= COMNSCOOKIE)) {
			// SOAP-ENV, SOAP-ENC, XSI, or COM
			strNamespace = m_mapNSShort[iNamespaceCookie];
		} // if
		else if ( ((unsigned long)iNamespaceCookie & (unsigned long)m_iBaseNSCookie) == m_iBaseNSCookie ) {
			// The cookie wasn't a standard namespace, so see if it
			// lies within the scope of this object. If the test above
			// passed, it does, so mask the scoping value and look
			// up the namespace.
			iNamespaceCookie &= 0x0000FFFF; // strip off scoping value
			strNamespace = m_mapNSShort[iNamespaceCookie];
		} // else if
		else {
			// The cookie wasn't a standard cookie, nor was it in
			// the scope of this element, so check to see if it
			// is registered with the parent node.
			if ( m_pParent != NULL ) {
				CSOAPElement* pNextParent = m_pParent->GetParent();
				if ( pNextParent != NULL ) {
					res = pNextParent->GetNamespace(iNamespaceCookie,strNamespace);
				} // if
				else {
					// The cookie must be invalid!
					res = SOAP_E_INVALIDARG; // invalid argument
				} // else
			} // if
			else {
				// The cookie must be invalid!
				res = SOAP_E_INVALIDARG; // invalid argument
			} // else
		} // else
	} // if

	return res;
}

unsigned long CSOAPElementNamespace::WriteToStream(std::ostrstream& ostr)
{
	// Nothing to write if the external stream is NULL
	if ( ostr == NULL ) return SOAP_E_POINTER;

	unsigned long res = SOAP_S_OK;
	try {
		std::string strNamespace;
		if ( m_bStdNamespaces[ENVNSCOOKIE] ) {
			// Stream the SOAP-ENV namespace
			strNamespace = " xmlns:";
			strNamespace += m_mapNSShort[ENVNSCOOKIE];
			strNamespace += "=\"";
			strNamespace += m_mapNSFull[ENVNSCOOKIE];
			strNamespace += "\"";
			ostr << strNamespace.c_str();
		} // if

		if ( m_bStdNamespaces[ENCNSCOOKIE] ) {
			// Stream the SOAP-ENC namespace
			strNamespace = " xmlns:";
			strNamespace += m_mapNSShort[ENCNSCOOKIE];
			strNamespace += "=\"";
			strNamespace += m_mapNSFull[ENCNSCOOKIE];
			strNamespace += "\"";
			ostr << strNamespace.c_str();
		} // if

		if ( m_bStdNamespaces[XSINSCOOKIE] ) {
			// Stream the XSI namespace
			strNamespace = " xmlns:";
			strNamespace += m_mapNSShort[XSINSCOOKIE];
			strNamespace += "=\"";
			strNamespace += m_mapNSFull[XSINSCOOKIE];
			strNamespace += "\"";
			ostr << strNamespace.c_str();
		} // if

		if ( m_bStdNamespaces[COMNSCOOKIE] ) {
			// Stream the COM namespace
			strNamespace = " xmlns:";
			strNamespace += m_mapNSShort[COMNSCOOKIE];
			strNamespace += "=\"";
			strNamespace += m_mapNSFull[COMNSCOOKIE];
			strNamespace += "\"";
			ostr << strNamespace.c_str();
		} // if

		// Stream any other namespaces
		for ( long i = COMNSCOOKIE+1; i < m_mapNSShort.size() ; i++ ) {
			strNamespace = " xmlns:";
			strNamespace += m_mapNSShort[i];
			strNamespace += "=\"";
			strNamespace += m_mapNSFull[i];
			strNamespace += "\"";

			// Write the namespace
			ostr << strNamespace.c_str();
		} // for
	} // try
	catch(unsigned long resErr) {
		// Some internal error...
		res = resErr; // normalize to this error...
		char szErr[256] = {0};
		sprintf(szErr,"Internal error while streaming element namespaces, %#08x.",resErr);
		std::cerr << szErr << std::endl;
	} // catch
	catch(...) {
		// Some error...
		res = SOAP_E_UNEXPECTED;
		std::cerr << "Unexpected application error while streaming element namespaces" << std::endl;
	} // catch

	return res;
}
