// SOAPEnvelope.cpp
//
// Kenn Scribner
// Understanding SOAP, March 2000
//
// SOAP Envelope object implementation
//

// Disable the "truncated to 255 characters" warning
// when expanding STL templates...
#pragma warning(disable : 4786)

#include "SOAPHeader.h"
#include "SOAPBody.h"
#include "SOAPEnvelope.h"

CSOAPEnvelope::CSOAPEnvelope(CSOAPHttpPacket* pParent) :
  CSOAPScopingElement(ENVELOPEBASECOOKIE),
  m_pParent(pParent),
  m_pHeader(NULL),
  m_pBody(NULL)
{
	// Hook up the attributes
	m_CAttributes.SetParent(this);

	// Establish tag text
	SetTagText("Envelope");
	SetTagNamespace(ENVNSCOOKIE);

	// Establish encoding
	EnableStdAttribute(ENCATTRCOOKIE,true);

	// Establish initial namespace list
	EnableStdNamespace(ENVNSCOOKIE,true);

	// Create the body...there will always be a
	// body (using normal SOAP encoding), but
	// delay creating the header until one is
	// requested.
	m_pBody = new CSOAPBody(this,NULL);
}

CSOAPEnvelope::~CSOAPEnvelope()
{
	// Delete the body
	delete m_pBody;
	m_pBody = NULL;

	// Delete the header (deleting a NULL pointer
	// is okay...).
	delete m_pHeader;
	m_pHeader = NULL;
}

CSOAPHeader* CSOAPEnvelope::GetHeader() {
	// If there is no SOAP header, create one
	// now...
	if ( m_pHeader == NULL ) {
		m_pHeader = new CSOAPHeader(this);
		m_pBody->SetHeader(m_pHeader);
	} // if

	return m_pHeader;
}

unsigned long CSOAPEnvelope::InsertElement(const std::string& strTagText, long iNamespaceCookie, CSOAPElement** ppElement)
{
	// You can't add independent elements to the envelope, so this
	// method simply returns a NULL element and SOAP_S_FALSE;
	if ( *ppElement == NULL ) return SOAP_E_POINTER;

	*ppElement = NULL;
	return SOAP_S_FALSE;
}

unsigned long CSOAPEnvelope::WriteToStream(std::ostrstream& ostr)
{
	// Nothing to write if the external stream is NULL
	if ( ostr == NULL ) return SOAP_E_POINTER;

	unsigned long res = SOAP_S_OK;
	try {
		// Write the opening envelope tag text to the
		// local stream
		res = StreamOpenTag(ostr);
		if ( res ) throw res;

		// Stream the header into the local stream, if
		// one has been created.
		if ( m_pHeader != NULL ) {
			res = m_pHeader->WriteToStream(ostr);
			if ( res ) throw res;
		} // if

		// Stream the body into the local stream
		res = m_pBody->WriteToStream(ostr);
		if ( res ) throw res;

		// Write the closing envelope tag text to the
		// local stream
		res = StreamCloseTag(ostr);
		if ( res ) throw res;
	} // try
	catch(unsigned long resErr) {
		// Some internal error...
		res = resErr; // normalize to this error...
		char szErr[256] = {0};
		sprintf(szErr,"Internal error while streaming envelope data, %#08x.",resErr);
		std::cerr << szErr << std::endl;
	} // catch
	catch(...) {
		// Some error...
		res = SOAP_E_UNEXPECTED;
		std::cerr << "Unexpected application error while streaming envelope data" << std::endl;
	} // catch

	return res;
}


unsigned long CSOAPEnvelope::StreamOpenTag(std::ostrstream& ostr)
{
	// Nothing to write if the external stream is NULL
	if ( ostr == NULL ) return SOAP_E_POINTER;

	unsigned long res = SOAP_S_OK;
	try {
		// Write the opening tag text to the local stream,
		// if the tag is named
		std::string sTag("<");

		// Stream namespace, if any
		std::string strNamespace;
		GetNamespace(m_iTagNamespaceCookie,strNamespace);
		sTag += strNamespace;

		// Concatenate the colon
		sTag += ":";

		// Concatenate the tag's element name
		sTag += m_strTagText;

		// Write the tag prior to attribute insertion
		ostr << sTag.c_str();

		// Stream any namespaces
		res = m_CNamespaces.WriteToStream(ostr);
		if ( res ) throw res;

		// Stream any attributes
		res = m_CAttributes.WriteToStream(ostr);
		if ( res ) throw res;

		// Terminate the tag
		sTag = ">";

		// Stream the tag
		ostr << sTag.c_str();
	} // try
	catch(unsigned long resErr) {
		// Some internal error...
		res = resErr; // normalize to this error...
		char szErr[256] = {0};
		sprintf(szErr,"Internal error while streaming envelope opening tag, %#08x.",resErr);
		std::cerr << szErr << std::endl;
	} // catch
	catch(...) {
		// Some error...
		res = SOAP_E_UNEXPECTED;
		std::cerr << "Unexpected application error while streaming envelope opening tag" << std::endl;
	} // catch

	return res;
}

unsigned long CSOAPEnvelope::StreamCloseTag(std::ostrstream& ostr)
{
	// Nothing to write if the external stream is NULL
	if ( ostr == NULL ) return SOAP_E_POINTER;

	unsigned long res = SOAP_S_OK;
	try {
		// Write the closing tag text to the local stream,
		// if there was data to enclose
		std::string sTag("</");

		// Stream namespace
		std::string strNamespace;
		GetNamespace(m_iTagNamespaceCookie,strNamespace);
		sTag += strNamespace;

		// Concatenate the colon
		sTag += ":";

		// Concatenate the tag's element name
		sTag += m_strTagText;

		// Terminate the tag
		sTag += ">";

		// Write the tag
		ostr << sTag.c_str();
	} // try
	catch(unsigned long resErr) {
		// Some internal error...
		res = resErr; // normalize to this error...
		char szErr[256] = {0};
		sprintf(szErr,"Internal error while streaming envelope closing tag, %#08x.",resErr);
		std::cerr << szErr << std::endl;
	} // catch
	catch(...) {
		// Some error...
		res = SOAP_E_UNEXPECTED;
		std::cerr << "Unexpected application error while streaming envelope closing tag" << std::endl;
	} // catch

	return res;
}
