// SOAPHttpPacket.cpp
//
// Kenn Scribner
// Understanding SOAP, March 2000
//
// SOAP HTTP object implementation
//

// Disable the "truncated to 255 characters" warning
// when expanding STL templates...
#pragma warning(disable : 4786)

#include "SOAPHttpPacket.h"

CSOAPHttpPacket::CSOAPHttpPacket(LPSOAPCALLINFO pci) :
  m_pEnvelope(NULL),
  m_iContentLength(0)
{
	// Create the envelope
	m_pEnvelope = new CSOAPEnvelope(this);

	// Copy in endpoint URL, POST information, and method name
	m_strEndpointURL = pci->strHost;
	m_strPostData = pci->strPostHdr;
	m_strActionData = pci->strActionData;
	m_strMethodName = pci->strMethod;

	// Create the POST header
	m_strPostText = " /"; // ' /InterfaceName HTTP/1.1'
	m_strPostText += m_strPostData;
	m_strPostText += " HTTP/1.1";

	m_strPostHdr = "POST"; // 'POST /InterfaceName HTTP/1.1<cr><lf>'
	m_strPostHdr += m_strPostText;
	m_strPostHdr += "\r\n";
	
	// Create the HOST header
	m_strHostText = "http://";
	if ( m_strEndpointURL.find(m_strHostText) == std::string::npos ) {
		// The "http://" needs to be appended
		m_strHostText += m_strEndpointURL;
	} // if
	else {
		// The "http://" is already part of the URL
		m_strHostText = m_strEndpointURL;
	} // else

	m_strHostHdr = "HOST: "; // 'Host: http://www.myurl.com<crt><lf>'
	m_strHostHdr += m_strHostText;
	m_strHostHdr += "\r\n";

	// Create Content-Type header
	m_strContentTypeText = "text/xml";
	m_strContentTypeHdr = "Content-Type: ";
	m_strContentTypeHdr += m_strContentTypeText;
	m_strContentTypeHdr += "\r\n";

	m_strContentLenText = ""; // will complete when streamed
	m_strContentLenHdr = "Content-Length: ";

	m_strSOAPActionText = m_strActionData;
	m_strSOAPActionHdr = "SOAPAction: \""; // 'SOAPAction: "MethodName"<cr><lf><cr><lf>'
	m_strSOAPActionHdr += m_strSOAPActionText;
	m_strSOAPActionHdr += "\"\r\n\r\n";
}

CSOAPHttpPacket::~CSOAPHttpPacket()
{
	// Delete the envelope
	delete m_pEnvelope;
	m_pEnvelope = NULL;
}

unsigned long CSOAPHttpPacket::WriteToStream(std::ostrstream& ostr, bool bEnvelopeOnly /*=false*/)
{
	// Nothing to write if the external stream is NULL
	if ( ostr == NULL ) return SOAP_E_POINTER;

	unsigned long res = SOAP_S_OK;
	try {
		if ( !bEnvelopeOnly ) {
			// The HTTP header information will come first as
			// the packet is sent to the remote site, but to make
			// the size calculations easier, the envelope is
			// serialized first. The HTTP information will then
			// be serialized such that it comes first with the
			// correct sizing.
			std::ostrstream ostrTemp;
			res = m_pEnvelope->WriteToStream(ostrTemp);
			if ( res ) throw res;

			// The overall packet size will be the size of the
			// SOAP XML information. 
			m_iContentLength = ostrTemp.pcount();

			// Copy in size
			char strLength[33] = {0}; // 32 digits plus terminator
			itoa(m_iContentLength,strLength,10);
			m_strContentLenHdr += strLength;
			m_strContentLenHdr += "\r\n";

			// With all of the header strings in place,
			// it's time to stream them out.
			//
			// Stream the header strings into the local stream
			ostr << m_strPostHdr.c_str();
			ostr << m_strHostHdr.c_str();
			ostr << m_strContentTypeHdr.c_str();
			ostr << m_strContentLenHdr.c_str();
			ostr << m_strSOAPActionHdr.c_str();
		} // if

		// Stream/restream the envelope
		res = m_pEnvelope->WriteToStream(ostr);
		if ( res ) throw res;
	} // try
	catch(unsigned long resErr) {
		// Some internal error...
		res = resErr; // normalize to this error...
		char szErr[256] = {0};
		sprintf(szErr,"Internal error while streaming HTTP packet, %#08x.",resErr);
		std::cerr << szErr << std::endl;
	} // catch
	catch(...) {
		// Some error...
		res = SOAP_E_UNEXPECTED;
		std::cerr << "Unexpected application error while streaming HTTP packet" << std::endl;
	} // catch

	return res;
}