// SOAPScopingElement.cpp
//
// Kenn Scribner
// Understanding SOAP, March 2000
//
// SOAP scoping element object implementation
//

// Disable the "truncated to 255 characters" warning
// when expanding STL templates...
#pragma warning(disable : 4786)

#include "SOAPScopingElement.h"
#include "SOAPIndElement.h"

CSOAPScopingElement::CSOAPScopingElement(long iBaseNSCookie, CSOAPElement* pParent /*=NULL*/) :
  CSOAPElement(pParent)
{
	// Establish the base cookie for namespace resolution
	m_CNamespaces.SetBaseNSCookie(iBaseNSCookie);
}

CSOAPScopingElement::~CSOAPScopingElement()
{
	// Free the independent elements
	for ( long i = 0; i < m_vecElements.size() ; i++ ) {
		// Delete the element
		CSOAPIndElement* pi = m_vecElements[i];
		delete pi;
	} // for
}

unsigned long CSOAPScopingElement::InsertElement(const std::string& strTagText, long iNamespaceCookie, CSOAPIndElement** ppElement)
{
	// A quick check
	if ( ppElement == NULL ) return SOAP_E_POINTER;

	// Create a new independent element
	CSOAPIndElement* pIE = new CSOAPIndElement(this);
	if ( pIE == NULL ) return SOAP_E_OUTOFMEMORY;

	// Establish its name and namespace
	pIE->SetTagText(strTagText);
	pIE->SetTagNamespace(iNamespaceCookie);

	// Assign out value
	*ppElement = pIE;

	// Stuff in the storage vector
	m_vecElements.push_back(pIE);

	return 0;
}

unsigned long CSOAPScopingElement::StreamContents(std::ostrstream& ostr)
{
	// Nothing to write if the external stream is NULL
	if ( ostr == NULL ) return SOAP_E_POINTER;

	unsigned long res = SOAP_S_OK;
	try {
		// NOTE: you can't have both internal data and
		// child nodes in SOAP, so prefer child nodes
		// over internal data...
		if ( !m_vecElements.size() ) {
			// Stream internal data
			res = CSOAPElement::StreamContents(ostr);
			if ( res ) throw res;
		} // if
		else {
			// Stream the contained elements
			for ( long i = 0; i < m_vecElements.size() ; i++ ) {
				// Retrieve the element
				CSOAPIndElement* pi = m_vecElements[i];
				res = pi->WriteToStream(ostr);
				if ( res ) throw res;
			} // for
		} // else
	} // try
	catch(unsigned long resErr) {
		// Some internal error...
		res = resErr; // normalize to this error...
		char szErr[256] = {0};
		sprintf(szErr,"Internal error while streaming scoping element contents, %#08x.",resErr);
		std::cerr << szErr << std::endl;
	} // catch
	catch(...) {
		// Some error...
		res = SOAP_E_UNEXPECTED;
		std::cerr << "Unexpected application error while streaming scoping element contents" << std::endl;
	} // catch

	return res;
}

unsigned long CSOAPScopingElement::StreamOpenTag(std::ostrstream& ostr)
{
	// Nothing to write if the external stream is NULL
	if ( ostr == NULL ) return SOAP_E_POINTER;

	unsigned long res = SOAP_S_OK;
	try {
		// Write the opening tag text to the local stream,
		// if the tag is named
		if ( m_strTagText.length() ) {
			std::string sTag("<");

			// Stream namespace, if any
			if ( m_iTagNamespaceCookie != NULLNSCOOKIE) {
				std::string strNamespace;
				GetNamespace(m_iTagNamespaceCookie,strNamespace);
				sTag += strNamespace;

				// Concatenate the colon
				sTag += ":";
			} // if

			// Concatenate the tag's element name
			sTag += m_strTagText;

			// Write the tag prior to attribute insertion
			ostr << sTag.c_str();

			// Stream any namespaces
			res = m_CNamespaces.WriteToStream(ostr);
			if ( res ) throw res;

			// Stream any attributes
			res = m_CAttributes.WriteToStream(ostr);
			if ( res ) throw res;

			// Terminate the tag
			if (m_strData.length() || m_vecElements.size()) {
				// Normal tag termination as this element has data
				// or independent elements.
				sTag = ">";
			} // if
			else {
				// Empty element tag termination, as this element has
				// no data (such as an embedded multi-reference
				// XPointer element in the method independent element).
				sTag = " />";
			} // else

			// Close the tag
			ostr << sTag.c_str();
		} // if
	} // try
	catch(unsigned long resErr) {
		// Some internal error...
		res = resErr; // normalize to this error...
		char szErr[256] = {0};
		sprintf(szErr,"Internal error while streaming scoping element opening tag, %#08x.",resErr);
		std::cerr << szErr << std::endl;
	} // catch
	catch(...) {
		// Some error...
		res = SOAP_E_UNEXPECTED;
		std::cerr << "Unexpected application error while streaming scoping element opening tag" << std::endl;
	} // catch

	return res;
}

unsigned long CSOAPScopingElement::StreamCloseTag(std::ostrstream& ostr)
{
	// Nothing to write if the external stream is NULL
	if ( ostr == NULL ) return SOAP_E_POINTER;

	unsigned long res = SOAP_S_OK;
	try {
		// Write the closing tag text to the local stream,
		// if there was data to enclose
		if (m_strData.length() || m_vecElements.size()) {
			std::string sTag("</");

			// Stream namespace, if any
			if ( m_iTagNamespaceCookie != NULLNSCOOKIE) {
				std::string strNamespace;
				GetNamespace(m_iTagNamespaceCookie,strNamespace);
				sTag += strNamespace;

				// Concatenate the colon
				sTag += ":";
			} // if

			// Concatenate the tag's element name
			sTag += m_strTagText;

			// Terminate the tag
			sTag += ">";

			// Write the tag
			ostr << sTag.c_str();
		} // if
	} // try
	catch(unsigned long resErr) {
		// Some internal error...
		res = resErr; // normalize to this error...
		char szErr[256] = {0};
		sprintf(szErr,"Internal error while streaming scoping element closing tag, %#08x.",resErr);
		std::cerr << szErr << std::endl;
	} // catch
	catch(...) {
		// Some error...
		res = SOAP_E_UNEXPECTED;
		std::cerr << "Unexpected application error while streaming scoping element closing tag" << std::endl;
	} // catch

	return res;
}

